import logging
from flask import request
from flask_jwt_extended import create_access_token, create_refresh_token, get_jwt_identity, \
    jwt_refresh_token_required, jwt_required
from flask_restplus import Namespace, Resource, fields

from portal import jwt
from portal.database.models import Component, User
from portal.service import general
from portal.service.auth import login_component, login_user
from portal.service.errors import PortalAPIError, UnauthorizedError

log = logging.getLogger(__name__)

auth_namespace = Namespace('auth')

login_schema = auth_namespace.model('LoginSchema', {
    'id': fields.String,
    'access_token': fields.String,
    'refresh_token': fields.String
    })

refresh_schema = auth_namespace.model('RefreshSchema', {
    'refresh_token': fields.String
    })

logout_schema = auth_namespace.model('LogoutSchema', {
    'access_token': fields.String,
    'refresh_token': fields.String
    })


@jwt.user_claims_loader
def add_claims_to_access_token(identity):
    data = 'user'
    if general.find_component(identity, throws=False):
        data = 'component'
    return dict(type=data)


def _login_user(data) -> User:
    username = data.get('username', None)
    password = data.get('password', None)
    gitlab_access_token = data.get('oauth_token', None)
    client = login_user(gitlab_access_token, password, username)
    return client


def _login_component(data) -> Component:
    name = data.get('name')
    if not name:
        raise PortalAPIError(400, message="Missing component name.")
    secret = data.get('secret')
    if not secret:
        raise PortalAPIError(400, message="Missing component secret.")
    sender_ip_address = request.remote_addr
    client = login_component(name, sender_ip_address, secret)
    return client


def _login_client(data: dict):
    if not data.get('type'):
        raise PortalAPIError(400, message="Missing login type.")

    login_type = data['type']
    types = dict(user=_login_user, component=_login_component)

    if login_type not in types.keys():
        raise PortalAPIError(400, message="Invalid login type.")

    return types[login_type](data)


def _create_tokens(client) -> dict:
    client_id = client.id
    return dict(
        id=client_id,
        access_token=create_access_token(identity=client_id),
        refresh_token=create_refresh_token(identity=client_id)
        )


@auth_namespace.route('/login')
class Login(Resource):
    @auth_namespace.marshal_with(login_schema)
    @auth_namespace.response(401, 'Client is not authorized')
    @auth_namespace.response(400, 'Missing login type.')
    @auth_namespace.response(400, 'Missing component name.')
    @auth_namespace.response(400, 'Missing component secret.')
    @auth_namespace.response(400, 'Invalid login type.')
    def post(self):
        data = request.get_json()
        client = _login_client(data)
        return _create_tokens(client=client)


def authorized_client():
    client = get_jwt_identity()
    if not client:
        raise UnauthorizedError()
    return client


@auth_namespace.route('/refresh')
class Refresh(Resource):

    @jwt_refresh_token_required
    @auth_namespace.marshal_with(refresh_schema)
    @auth_namespace.response(401, 'Client is not authorized')
    def post(self):
        client = authorized_client()
        return dict(access_token=create_access_token(identity=client))


@auth_namespace.route('/logout')
class Logout(Resource):

    @jwt_required
    @auth_namespace.marshal_with(logout_schema)
    @auth_namespace.response(401, 'Client is not authorized')
    def post(self):
        authorized_client()
        return dict(access_token=None, refresh_token=None)
