import json
import logging
from flask import request
from flask_jwt_extended import jwt_required
from flask_restplus import Namespace, Resource

from portal.database.models import ProjectState
from portal.rest import rest_helpers
from portal.rest.schemas import config_schema, config_schema_reduced, project_schema, \
    projects_schema, submission_create_schema, submission_schema, submissions_schema
from portal.service import auth, general
from portal.service.errors import ForbiddenError, SubmissionRefusedError
from portal.service.permissions import check_client, require_client
from portal.service.projects import can_create_submission, create_project, delete_project, \
    find_project_submissions, list_projects, update_project, update_project_config
from portal.service.submissions import create_submission
from portal.tools import time

projects_namespace = Namespace('')

log = logging.getLogger(__name__)


@projects_namespace.route('/courses/<string:cid>/projects')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.response(404, 'Course not found')
class ProjectsList(Resource):

    @jwt_required
    @projects_namespace.response(404, 'Course not found')
    #@projects_namespace.response(200, 'Projects list', model=projects_schema)
    def get(self, cid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        return projects_schema.dump(list_projects(course, client))

    @jwt_required
    @projects_namespace.response(404, 'Course not found')
    #@projects_namespace.response(201, 'Project created', model=project_schema)
    def post(self, cid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        perm = ['update_course']
        require_client(client=client, course=course, permissions=perm)

        data = rest_helpers.parse_request_data(
            project_schema, action='create', resource='project'
            )

        new_project = create_project(course, data)
        return project_schema.dump(new_project)[0], 201


@projects_namespace.route('/courses/<string:cid>/projects/<string:pid>')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.param('pid', 'Project id')
@projects_namespace.response(404, 'Course not found')
@projects_namespace.response(404, 'Project not found')
class ProjectResource(Resource):
    @jwt_required
    #@projects_namespace.response(200, 'Project found', model=project_schema)
    @projects_namespace.response(403, 'Not allowed to get project')
    def get(self, cid: str, pid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        perm = ['view_course_full', 'view_course_limited']
        require_client(client=client, course=course, permissions=perm)

        project = general.find_project(course, pid)
        return project_schema.dump(project)

    @jwt_required
    @projects_namespace.response(204, 'Project deleted')
    @projects_namespace.response(403, 'Not allowed to delete project')
    def delete(self, cid: str, pid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        perm = ['update_course']
        require_client(client=client, course=course, permissions=perm)

        project = general.find_project(course, pid)
        delete_project(project)
        return '', 204

    @jwt_required
    @projects_namespace.response(204, 'Project updated')
    @projects_namespace.response(403, 'Not allowed to update project')
    def put(self, cid: str, pid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        perm = ['write_projects', 'update_course']
        require_client(client=client, course=course, permissions=perm)

        data = rest_helpers.parse_request_data(
            project_schema, action='update', resource='project'
            )

        project = general.find_project(course, pid)
        update_project(project, data)
        return '', 204


@projects_namespace.route('/courses/<string:cid>/projects/<string:pid>/config')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.param('pid', 'Project id')
@projects_namespace.response(404, 'Course not found')
@projects_namespace.response(404, 'Project not found')
class ProjectConfigResource(Resource):
    @jwt_required
    #@projects_namespace.response(200, 'Config of the project', model=config_schema)
    #@projects_namespace.response(200, 'Config of the project', model=config_schema_reduced)
    def get(self, cid: str, pid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        project = general.find_project(course, pid)
        # authorization
        schema = get_config_schema_based_on_permissions(client, course)
        return schema.dump(project.config)

    @jwt_required
    @projects_namespace.response(204, 'Project config updated')
    def put(self, cid: str, pid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        perm = ['write_projects', 'update_course']
        require_client(client=client, course=course, permissions=perm)

        data = rest_helpers.parse_request_data(
            config_schema, action='update', resource='project_config'
            )

        project = general.find_project(course, pid)
        update_project_config(project, data)
        return '', 204


@projects_namespace.route('/courses/<string:cid>/projects/<string:pid>/submissions')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.param('pid', 'Project id')
@projects_namespace.response(404, 'Course not found')
@projects_namespace.response(404, 'Project not found')
class ProjectSubmissions(Resource):
    @jwt_required
    #@projects_namespace.response(200, 'Project submissions list', model=submissions_schema)
    def get(self, cid: str, pid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        perm = ['read_submissions_all']
        require_client(client=client, course=course, permissions=perm)

        user_id = request.args.get('user')
        project = general.find_project(course, pid)
        submissions = find_project_submissions(project, user_id)
        return submissions_schema.dump(submissions)

    @jwt_required
    #@projects_namespace.response(201, 'Project submission create', model=submission_schema)
    def post(self, cid: str, pid: str):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        perm = ['create_submissions']
        require_client(client=client, course=course, permissions=perm)

        # check if a new submission can be created by this user in this project
        project = general.find_project(course, pid)
        _check_submission_create(client, project)

        data = rest_helpers.parse_request_data(
            submission_create_schema, action='create', resource='submission'
            )

        # data for Kontr processing
        project_params = data.get('project_params')
        project_params_string = json.dumps(project_params) or ""

        new_submission = create_submission(
            user=client,
            project=project,
            file_params=data['file_params'],
            project_params_string=project_params_string
            )
        return submission_schema.dump(new_submission)[0], 201


@projects_namespace.route('/courses/<string:cid>/projects/<string:pid>/files')
@projects_namespace.param('cid', 'Course id')
@projects_namespace.param('pid', 'Project id')
@projects_namespace.response(404, 'Course not found')
@projects_namespace.response(404, 'Project not found')
class ProjectTestFiles(Resource):

    def get(self, cid: str, pid: str):
        # Gets test files (feature)
        raise NotImplementedError()


def _check_submission_create(client, project):
    if project.state(timestamp=time.NOW()) != ProjectState.ACTIVE:
        raise SubmissionRefusedError(f"Project {project.name} not active.")
    if not can_create_submission(client, project):
        raise SubmissionRefusedError(f"Submission in project {project.name} already active.")


def get_config_schema_based_on_permissions(client, course):
    if check_client(client=client, course=course, permissions=['view_course_full']):
        return config_schema
    elif check_client(client=client, course=course, permissions=['view_course_limited']):
        return config_schema_reduced
    else:
        raise ForbiddenError(uid=client.id)
