"""
Errors in the service layer
"""
import json
from typing import Union


class PortalError(Exception):
    """Base exception class
    """


class PortalConfigurationLoadError(PortalError):
    """Error thrown when configuration is missing or is invalid
    """
    pass


class PortalAPIError(PortalError):
    """Portal API Error thrown when error in the API happened
    """

    def __init__(self, code: int, message: Union[str, dict]):
        """Creates instance of the API portal error
        Args:
            code(int): Status code
            message(Union[str, dict]): Message
        """
        self._code = code
        self._message = message

    @property
    def code(self) -> int:
        """Gets the status code that should be returned
        Returns(int): Status code

        """
        return self._code

    @property
    def message(self) -> Union[str, dict]:
        """Gets the error message
        Returns(Union[str, dict]): Message
        """
        return self._message

    @property
    def json(self):
        """Json serialized message
        Returns:

        """
        return json.dumps(self.message)

    def __str__(self):
        return self.__class__.__name__ + ': ' + self.json


class DataMissingError(PortalAPIError):
    """Data missing error thrown when data is missing i.e. (update)
    """

    def __init__(self, action: str, resource: str):
        """Creates data missing error instance
        Args:
            action(str): Name of the action
            resource(str): Name of the resource
        """
        message = dict(action=action, resource=resource, message='Data is missing!')
        super(DataMissingError, self).__init__(code=400, message=message)


class ResourceNotFoundError(PortalAPIError):
    """Resource not found error
    """

    def __init__(self, resource: str, res_id=None, note=None):
        """Creates instance of the error
        Args:
            resource(str): Name of the resource
            res_id(str): UUID of the resource
            note(str): Note
        """
        message = dict(
            resource=resource,
            id=res_id,
            message=f"Cannot find {resource} with identifier: {res_id}",
            )
        if note:
            message['note'] = note

        super().__init__(code=404, message=message)


class UnauthorizedError(PortalAPIError):
    def __init__(self, note=None):
        message = dict(
            message=f"You are not authorized.",
            )
        if note:
            message['note'] = note

        super().__init__(code=401, message=message)


class ForbiddenError(PortalAPIError):
    # could use a resource identification (like 404)
    def __init__(self, uid=None, note=None):
        user_message = f"Forbidden for user: {uid}!" if uid else "Forbidden action."
        message = dict(
            uid=uid,
            message=user_message,
            )
        if note:
            message['note'] = note

        super().__init__(code=403, message=message)


class IncorrectCredentialsError(UnauthorizedError):
    def __init__(self):
        super().__init__(note="Incorrect username or password.")


class InvalidGitlabAccessTokenError(UnauthorizedError):
    def __init__(self):
        super().__init__(note='Invalid Gitlab Access token.')


class ValidationError(PortalAPIError):  # currently not used
    def __init__(self, schema, data, failures: list):
        message = dict(
            schema=schema,
            data=data,
            failures=failures,
            )

        super().__init__(code=400, message=message)


class SubmissionRefusedError(PortalAPIError):
    def __init__(self, reason):
        super().__init__(code=400, message=reason)

