"""
Projects service
"""

import logging
from flask_sqlalchemy import BaseQuery
from typing import List

from portal.database.models import Course, Project, ProjectConfig, Submission, SubmissionState, User
from portal.service import filters, general
from portal.service.errors import ForbiddenError
from portal.service.general import get_new_name
from portal.service.permissions import check_client

log = logging.getLogger(__name__)


def copy_project(source: Project, target: Course) -> Project:
    """Copies a project to the target course

    Args:
        source(Project): Project instance
        target(Course): Course instance

    Returns(Project): Copied project
    """
    new_name = get_new_name(source, target)
    new_project = Project(target, new_name, source.config.test_files_source)
    new_project.set_config(**vars(source.config))
    return new_project


def can_create_submission(user: User, project: Project) -> bool:
    """Checks if a user may create a new submission.

    A user is permitted to create a new submission if there are no submissions
    created or queued for him in the given project.

    Args:
        user(User): the user who wants to crete a new submission
        project(Project): the project in which the user wants to create a submission

    Returns: whether the user can create a new submission in the project

    """
    conflicting_states = [SubmissionState.CREATED, SubmissionState.READY, SubmissionState.QUEUED]
    conflict = Submission.query.filter(Submission.user == user) \
        .filter(Submission.project == project) \
        .filter(Submission.state.in_(conflicting_states)) \
        .first()

    return conflict is None


def delete_project(project: Project):
    """Deletes a project

    Args:
        project(Project): project instance
    """
    log.info(f"[DELETE] Project {project.id} ({project.name}).")
    general.delete_entity(project)


def update_project(project: Project, data: dict) -> Project:
    """Updates a project

    Args:
        project(Project): Project instance
        data(dict): Data dictionary

    Returns(Project): Updated project
    """
    log.info(f"[UPDATE] Project {project.id} ({project.name}): {data}.")
    return __set_project_data(project, data)


def __set_project_data(project: Project, data: dict) -> Project:
    return general.update_entity(project, data, allowed=['name'])


def create_project(course: Course, data: dict) -> Project:
    """Creates a new project

    Args:
        course(Course): Course instance
        data(dict): Project data

    Returns(Project): Project instance
    """
    test_files_source = data['config']['test_files_source']
    new_project = Project(course=course, test_files_source=test_files_source)
    __set_project_data(project=new_project, data=data)
    log.info(f"[CREATE] Project for course {course.id} ({course.codename}): "
             f"{new_project}")
    return new_project


def update_project_config(project: Project, data: dict) -> ProjectConfig:
    """Updates project config

    Args:
        project(Project): Project instance
        data(dict): Config data

    Returns(ProjectConfig): Project config instance
    """
    course = project.course
    project.set_config(**data)
    general.write_entity(project)
    log.info(f"[UPDATE] Configuration for project {project.id} in '"
             f"course {course.id} to {data}.")
    return project.config


def query_submissions(project: Project, user: User) -> BaseQuery:
    """Queries submissions for the project and filters them by user.

    Args:
        project(Project): Project instance
        user(User): User instance
    """
    course = project.course
    return Submission.query.filter(Submission.user_id == user.id) \
        .filter(Submission.project_id == project.id) \
        .join(Course, Course.id == project.course_id).filter(Course.id == course.id)


def find_project_submissions(project: Project, user_id: str = None) -> List[Submission]:
    """Finds all project submissions. If a user id is specified, returns only submissions created
    by that user.

    Args:
        project(Project): Project instance
        user_id(str): User id (optional)

    Returns(List[Submission]): A list of all submissions in a project if no user_id is specified.
    Otherwise a list of submissions created in the project by the specified user.

    """
    submissions = project.submissions
    if user_id:
        user = general.find_user(user_id)
        submissions = query_submissions(project, user)
    return submissions


def list_projects(course: Course, client) -> list:
    """List of all projects
    Args:
        course(Course): Course instance
        client: Client instance

    Returns(list): list of projects

    """
    if check_client(client=client, course=course, permissions=['view_course_full']):
        return course.projects
    elif check_client(client=client, course=course, permissions=['view_course_limited']):
        return filters.filter_projects_from_course(course=course, user=client)
    raise ForbiddenError(uid=client.id)
